﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class Crane : MonoBehaviour
{
    public Transform transformBasestation   = null;
    public Transform transformClaw          = null;
    public Transform transformGrabLocator   = null;

    public Transform transformClawFork1     = null;
    public Transform transformClawFork2     = null;
    public Transform transformClawFork3     = null;
    public Transform transformClawFork4     = null;


    // tuning parameters
    private readonly Vector3 START_POSITION = new Vector3(0, 2.25f, -1f);
    private const float MOVEMENT_SPEED      = 1.5f*1.5f;
    private const float CLAW_SPEED          = 2.0f*1.25f;
    private const float GRAB_Y              = 0.7f;
    private const float CLAW_MOVE_BACK_AT_Y = 0.2f;
    private const float LOOT_FALLACCELL     = 9.81f;

    private const float CLAW_OPEN_ANGLE     = 300f;
    private const float CLAW_CLOSED_ANGLE   = 345f;
    private const float CLAW_GRAB_SPEED     = 3.5f;
    private const float GRAB_RANGE          = 1f;

    public enum State
    {
        AWAITING_INPUT,
        MOVING_TO_TARGET_HORIZONTAL,
        MOVING_TO_TARGET_VERTICAL,
        CLOSE_CLAW,
        MOVING_BACK_VERTICAL,
        MOVING_BACK_HORIZONTAL,
        OPEN_CLAW,
        LOOT_FALL,
    }
    private State state = State.AWAITING_INPUT;
    private Vector3 targetPosition = Vector3.zero;
    private GameObject potentialLoot = null;
    private GameObject loot = null;

    private Transform transformLootAnimation = null;

    private bool isPlayerInputEnabled = false;
    private bool isUpdatedByServer = false;
    // Network related
    private Vector3 targetPositionFromServer;
    private bool isRequestedInputFromServer;

    private bool caughtSnake = false;
    private int numMoves = 0;
    private Vector3 lootFallVelocity = Vector3.zero;

    public GameObject GetAndRemoveLootObject()
    {
        GameObject tmpLoot = loot;
        loot = null;
        return tmpLoot;
    }

    public int GetAndResetNumMoves()
    {
        int tmpNumMoves = numMoves;
        numMoves = 0;
        return tmpNumMoves;
    }


    public void EnablePlayerInput()
    {
        // Input should be sent to the server
        isPlayerInputEnabled = true;
    }
    public void SetIsUpdatedByServer(bool isUpdatedByServer)
    {
        this.isUpdatedByServer = isUpdatedByServer;
    }



    private void Awake()
    {
        transformBasestation.position = START_POSITION;
        transformClaw.position = START_POSITION;
        transformLootAnimation = new GameObject("LootLocator").transform;
        transformLootAnimation.SetParent(transformGrabLocator);
        ResetLootLocator();
    }

    private void ResetLootLocator()
    {
        transformLootAnimation.localPosition = Vector3.zero;
        transformLootAnimation.localScale = Vector3.one;
        transformLootAnimation.localRotation = Quaternion.identity;
    }

    public State GetState()
    {
        return state;
    }



    // Update is called once per frame
    void Update()
    {
        if(isPlayerInputEnabled)
        {
            // only send input to server
            if (HasNewTargetPositionLocal(out Vector3 tmpTargetPosition))
            {
                SendTargetPositionToServer(tmpTargetPosition);
            }
            if (isUpdatedByServer)
            {
                return;
            }
        }

        // Server
        if (state == State.AWAITING_INPUT)
        {
            if (HasNewTargetPositionRemote(out targetPosition))
            {
                isRequestedInputFromServer = false;
                ChangeState(State.MOVING_TO_TARGET_HORIZONTAL);
            }
        }
        else if (state == State.MOVING_TO_TARGET_HORIZONTAL)
        {

            Vector3 targetPositionBasestation = new Vector3(targetPosition.x, transformBasestation.position.y, targetPosition.z);

            Vector3 vectorMovement = targetPositionBasestation - transformBasestation.position;
            Vector3 velocity = vectorMovement.normalized * MOVEMENT_SPEED * Time.deltaTime;
            velocity = velocity.sqrMagnitude > vectorMovement.sqrMagnitude ? vectorMovement : velocity;

            transformBasestation.position = transformBasestation.position + velocity;
            transformClaw.position = new Vector3(transformBasestation.position.x, transformClaw.position.y, transformBasestation.position.z);


            if (velocity.sqrMagnitude < Mathf.Epsilon)
            {
                ChangeState(State.MOVING_TO_TARGET_VERTICAL);
            }
        }
        else if (state == State.MOVING_TO_TARGET_VERTICAL)
        {
            transformClaw.position = new Vector3(transformClaw.position.x,
                                                    Mathf.Lerp(transformClaw.position.y, targetPosition.y, CLAW_SPEED * Time.deltaTime),
                                                    transformClaw.position.z);

            if (transformClaw.position.y < GRAB_Y)
            {
                ChangeState(State.CLOSE_CLAW);
            }
        }
        else if (state == State.CLOSE_CLAW)
        {

            Vector3 l_euler = transformClawFork1.transform.rotation.eulerAngles;
            l_euler.x = Mathf.Lerp(l_euler.x, CLAW_CLOSED_ANGLE, CLAW_GRAB_SPEED * Time.deltaTime);
            transformClawFork1.transform.rotation = Quaternion.Euler(l_euler);

            l_euler = transformClawFork2.transform.rotation.eulerAngles;
            l_euler.x = Mathf.Lerp(l_euler.x, CLAW_CLOSED_ANGLE, CLAW_GRAB_SPEED * Time.deltaTime);
            transformClawFork2.transform.rotation = Quaternion.Euler(l_euler);

            l_euler = transformClawFork3.transform.rotation.eulerAngles;
            l_euler.x = Mathf.Lerp(l_euler.x, CLAW_CLOSED_ANGLE, CLAW_GRAB_SPEED * Time.deltaTime);
            transformClawFork3.transform.rotation = Quaternion.Euler(l_euler);

            l_euler = transformClawFork4.transform.rotation.eulerAngles;
            l_euler.x = Mathf.Lerp(l_euler.x, CLAW_CLOSED_ANGLE, CLAW_GRAB_SPEED * Time.deltaTime);
            transformClawFork4.transform.rotation = Quaternion.Euler(l_euler);

            if (Mathf.Abs(CLAW_CLOSED_ANGLE - l_euler.x) < 1f)
            {
                ChangeState(State.MOVING_BACK_VERTICAL);
            }
        }
        else if (state == State.MOVING_BACK_VERTICAL)
        {
            transformClaw.position = new Vector3(transformClaw.position.x,
                                                    Mathf.Lerp(transformClaw.position.y, START_POSITION.y, CLAW_SPEED * Time.deltaTime),
                                                    transformClaw.position.z);


            if (Mathf.Abs(START_POSITION.y - transformClaw.position.y) < CLAW_MOVE_BACK_AT_Y)
            {
                Vector3 targetPositionBasestation = START_POSITION;

                Vector3 vectorMovement = targetPositionBasestation - transformBasestation.position;
                Vector3 velocity = vectorMovement.normalized * MOVEMENT_SPEED * Time.deltaTime;
                velocity = velocity.sqrMagnitude > vectorMovement.sqrMagnitude ? vectorMovement : velocity;

                transformBasestation.position = transformBasestation.position + velocity;
                transformClaw.position = new Vector3(transformBasestation.position.x, transformClaw.position.y, transformBasestation.position.z);


                if (velocity.sqrMagnitude < Mathf.Epsilon)
                {
                    ChangeState(State.OPEN_CLAW);
                }
            }
        }
        else if (state == State.MOVING_BACK_HORIZONTAL)
        {

        }
        else if (state == State.OPEN_CLAW)
        {
            Vector3 l_euler = transformClawFork1.transform.rotation.eulerAngles;
            l_euler.x = Mathf.Lerp(l_euler.x, CLAW_OPEN_ANGLE, CLAW_GRAB_SPEED * Time.deltaTime);
            transformClawFork1.transform.rotation = Quaternion.Euler(l_euler);

            l_euler = transformClawFork2.transform.rotation.eulerAngles;
            l_euler.x = Mathf.Lerp(l_euler.x, CLAW_OPEN_ANGLE, CLAW_GRAB_SPEED * Time.deltaTime);
            transformClawFork2.transform.rotation = Quaternion.Euler(l_euler);

            l_euler = transformClawFork3.transform.rotation.eulerAngles;
            l_euler.x = Mathf.Lerp(l_euler.x, CLAW_OPEN_ANGLE, CLAW_GRAB_SPEED * Time.deltaTime);
            transformClawFork3.transform.rotation = Quaternion.Euler(l_euler);

            l_euler = transformClawFork4.transform.rotation.eulerAngles;
            l_euler.x = Mathf.Lerp(l_euler.x, CLAW_OPEN_ANGLE, CLAW_GRAB_SPEED * Time.deltaTime);
            transformClawFork4.transform.rotation = Quaternion.Euler(l_euler);


            lootFallVelocity += Vector3.down * LOOT_FALLACCELL * Time.deltaTime;
            transformLootAnimation.transform.position = transformLootAnimation.transform.position + lootFallVelocity * Time.deltaTime;
            if (transformLootAnimation.transform.position.y < -15f)
            {
                transformLootAnimation.transform.position = new Vector3(transformLootAnimation.transform.position.x, -15, transformLootAnimation.transform.position.z);
            }
            if (Mathf.Abs(CLAW_OPEN_ANGLE - l_euler.x) < 1f)
            {
                ChangeState(State.AWAITING_INPUT);
            }
        }
        else if (state == State.LOOT_FALL)
        {
            
        }
    }


    private void ChangeState(State newState)
    {
        if (state == State.AWAITING_INPUT)
        {
        }
        else if (state == State.MOVING_TO_TARGET_HORIZONTAL)
        {
        }
        else if (state == State.MOVING_TO_TARGET_VERTICAL)
        {
        }
        else if (state == State.CLOSE_CLAW)
        {
            float sqrDistanceClosest = GRAB_RANGE * GRAB_RANGE;
            // Check snake first
            GameObject objectClosest = null;
            Snake snake = GameObject.FindObjectOfType<Snake>();
            float sqrDistanceSnake = (snake.transform.position - transformGrabLocator.position).sqrMagnitude;
            if (sqrDistanceSnake < sqrDistanceClosest)
            {
                objectClosest = snake.gameObject;
                sqrDistanceClosest = sqrDistanceSnake;
            }

            foreach (Ball ball in GameObject.FindObjectsOfType<Ball>())
            {
                float sqrDistanceCurrent = (ball.transform.position - transformGrabLocator.position).sqrMagnitude;
                if (sqrDistanceCurrent < sqrDistanceClosest)
                {
                    objectClosest = ball.gameObject;
                    sqrDistanceClosest = sqrDistanceCurrent;
                }
            }

            if (objectClosest != null)
            {
                if (snake.gameObject == objectClosest)
                {
                    snake.transform.SetParent(transformLootAnimation);
                    snake.transform.localPosition = Vector3.zero;
                    potentialLoot = snake.gameObject;
                }
                else
                {
                    Ball ball = objectClosest.GetComponent<Ball>();
                    if (snake.HasBall(ball))
                    {
                        snake.Split(ball, transformLootAnimation);
                    }
                    else //if( ball.IsOnFloor() )
                    {
                        ball.DisablePhysics();
                        ball.transform.SetParent(transformLootAnimation);
                        ball.transform.localPosition = Vector3.zero;
                    }
                    potentialLoot = ball.gameObject;
                }
            }
        }
        else if (state == State.MOVING_BACK_VERTICAL)
        {
        }
        else if (state == State.MOVING_BACK_HORIZONTAL)
        {
        }
        else if (state == State.OPEN_CLAW)
        {
            if (potentialLoot != null)
            {
                potentialLoot.transform.SetParent(null);
                loot = potentialLoot;
                lootFallVelocity = Vector3.zero;
                transformLootAnimation.localPosition = Vector3.zero;
            }
        }
        else if (state == State.LOOT_FALL)
        {
        }


        if (newState == State.AWAITING_INPUT)
        {
        }
        else if (newState == State.MOVING_TO_TARGET_HORIZONTAL)
        {
            numMoves++;
        }
        else if (newState == State.MOVING_TO_TARGET_VERTICAL)
        {
        }
        else if (newState == State.CLOSE_CLAW)
        {
        }
        else if (newState == State.MOVING_BACK_VERTICAL)
        {
        }
        else if (newState == State.MOVING_BACK_HORIZONTAL)
        {
        }
        else if (newState == State.OPEN_CLAW)
        {
        }
        else if (newState == State.LOOT_FALL)
        {
        }
        state = newState;
    }


    public void OpenClawDirectly()
    {
        SetClawOpenState(1);
    }

    public void CloseClawDirectly()
    {
        SetClawOpenState(0);
    }

    private void SetClawOpenState(float time01)
    {
        Vector3 l_euler = transformClawFork1.transform.rotation.eulerAngles;
        l_euler.x = Mathf.Lerp(CLAW_CLOSED_ANGLE, CLAW_OPEN_ANGLE, time01);
        transformClawFork1.transform.rotation = Quaternion.Euler(l_euler);

        l_euler = transformClawFork2.transform.rotation.eulerAngles;
        l_euler.x = Mathf.Lerp(CLAW_CLOSED_ANGLE, CLAW_OPEN_ANGLE, time01);
        transformClawFork2.transform.rotation = Quaternion.Euler(l_euler);

        l_euler = transformClawFork3.transform.rotation.eulerAngles;
        l_euler.x = Mathf.Lerp(CLAW_CLOSED_ANGLE, CLAW_OPEN_ANGLE, time01);
        transformClawFork3.transform.rotation = Quaternion.Euler(l_euler);

        l_euler = transformClawFork4.transform.rotation.eulerAngles;
        l_euler.x = Mathf.Lerp(CLAW_CLOSED_ANGLE, CLAW_OPEN_ANGLE, time01);
        transformClawFork4.transform.rotation = Quaternion.Euler(l_euler);
    }




    // NETWORK

    public void SetTargetInputFromServer(Vector3 targetPosition)
    {
        targetPositionFromServer = targetPosition;
        isRequestedInputFromServer = true;
    }


    public void SendTargetPositionToServer(Vector3 targetPosition)
    {
        CranePhoton cranePhoton = GetComponent<CranePhoton>();
        if (cranePhoton != null)
        {
            cranePhoton.photonView.RPC("SetTargetPosition", Photon.Pun.RpcTarget.All, targetPosition);
        }
        else
        {
            // Playing local
            SetTargetInputFromServer(targetPosition);
        }
    }


    private bool HasNewTargetPositionRemote(out Vector3 oTargetPosition)
    {
        oTargetPosition = targetPositionFromServer;
        return isRequestedInputFromServer;
    }


    private bool HasNewTargetPositionLocal(out Vector3 oTargetPosition)
    {
        if (Input.GetMouseButtonUp(0))
        {
            Ray ray = Camera.main.ScreenPointToRay(Input.mousePosition);
            // Save the info
            RaycastHit hit;
            // You successfully hi
            if (Physics.Raycast(ray, out hit, float.PositiveInfinity, LayerMask.GetMask("CraneOperator")))
            {
                oTargetPosition = hit.point;
                return true;
            }
        }
        oTargetPosition = Vector3.zero;
        return false;
    }

}
