﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class Snake : MonoBehaviour
{
    public static int NumBallsDestroyedBySnake = 0;
    private class SnakeBall
    {
        private enum State
        {
            INITIAL,
            DIGESTING,
            GROWING,
            ADDED
        }

        public Transform parent = null;
        private State state = State.INITIAL;
        public const float DIGESTION_DURATION = 0.1f;
        public const float GROWING_DURATION = 0.5f;
        public const float DISTANCE_TO_PARENT = 0.15f; //important tuning parameter, depends on ball size
        public const float SNAKE_BALL_SIZE_FACTOR = 0.875f * 1.0f; //larger than 1 felt bulky, but let's see what happens when we put the snake art in. Perhaps same size collider but larger mesh for the snake?

        public Ball ball;
        public float growingTimeLeft = GROWING_DURATION;
        public float digestionTimeLeft = DIGESTION_DURATION;
        public Queue<Vector3> previousParentPositions = new Queue<Vector3>();

        public SnakeBall(Ball ball, Transform parent)
        {
            this.ball = ball;
            this.parent = parent;
            ChangeState(State.DIGESTING);
        }

        public void Update()
        {
            if (state == State.DIGESTING)
            {
                digestionTimeLeft -= Time.deltaTime;
                if (digestionTimeLeft < 0f)
                {
                    ChangeState(State.GROWING);
                }
            }
            else if (state == State.GROWING)
            {
                growingTimeLeft -= Time.deltaTime;

                float t = 1f - (growingTimeLeft / GROWING_DURATION);
                float growingDiameter = Mathf.Lerp(0f, ball.GetOriginalDiameter()* SNAKE_BALL_SIZE_FACTOR, t);
                ball.transform.localScale = new Vector3(growingDiameter, growingDiameter, growingDiameter);

                if (growingTimeLeft < 0f)
                {
                    ChangeState(State.ADDED);
                }

                FollowParent();
            }
            else if (state == State.ADDED)
            {
                FollowParent();
            }
        }

        private void FollowParent()
        {
            if (parent)
            {
                previousParentPositions.Enqueue(parent.position);

                while (previousParentPositions.Count > 0 && (parent.position - previousParentPositions.Peek()).sqrMagnitude > DISTANCE_TO_PARENT * DISTANCE_TO_PARENT)
                {
                    ball.transform.position = previousParentPositions.Dequeue();
                }

                // set at precise distance to parent
                ball.transform.position = parent.transform.position - (parent.transform.position - ball.transform.position).normalized * DISTANCE_TO_PARENT;
            }
        }

        private void ChangeState(State newState)
        {
            //Debug.Log("SnakeBall.ChangeState called for ball " +  ball.transform.name + " , newState: " + newState);

            // exit old state
            if (state == State.INITIAL)
            {

            }
            else if (state == State.DIGESTING)
            {

            }
            else if (state == State.GROWING)
            {
                
            }
            else if (state == State.ADDED)
            {

            }

            // init new state
            if (newState == State.DIGESTING)
            {
                ball.transform.localScale = Vector3.zero;

                //NOTE this is never set back to non-kinematic because a severed part of the snake is either removed or moved around and removed later
                ball.GetComponent<Rigidbody>().isKinematic = true;
            }
            else if (newState == State.GROWING)
            {
                if (parent)
                {
                    //spawns the ball at rear end back of the previous one (without any offset)
                    ball.transform.position = parent.position - parent.forward * (parent.localScale.x / 2f + ball.transform.localScale.x / 2f);
                }
                else
                {
                    Debug.LogWarning("Snakeball is still growing and it had no parent - has the player managed to grab a growing ball at the very end of the snake?");
                }
            }
            else if (newState == State.ADDED)
            {

            }

            state = newState;
        }

        public void Destroy()
        {
            GameObject.Destroy(ball.gameObject);
            Snake.NumBallsDestroyedBySnake++;
        }

        public bool IsGrown()
        {
            return (state == State.ADDED);
        }
    }

    private class SnakePart
    {
        public List<SnakeBall> snakeBalls = new List<SnakeBall>();

        public SnakePart(List<SnakeBall> snakeBalls, Transform newParent)
        {
            this.snakeBalls = snakeBalls;

            if (snakeBalls.Count > 0)
            {
                snakeBalls[0].parent = newParent;
            }
            else
            {
                Debug.LogWarning("SnakePart constructor called but with empty collection of snakeBalls.");
            }
        }

        public void Update()
        {
            foreach (SnakeBall snakeBall in snakeBalls)
            {
                snakeBall.Update();
            }
        }

        public void Destroy()
        {
            foreach (SnakeBall snakeBall in snakeBalls)
            {
                snakeBall.Destroy();
            }
            snakeBalls.Clear();
        }
    }

    private List<SnakeBall> snakeBalls = new List<SnakeBall>();
    private Ball lastEdibleBallNearMouth = null;
    private Vector3 movementDirection;
    private float SPEED = 0.75f; // 0.25f was way too slow. Perhaps increase speed for a certain amount of time after clicking, and with additional speed boosts each time you pick up a ball?
    private List<SnakePart> snakeParts = new List<SnakePart>();
    private bool isPlayerInputEnabled = false;
    private bool isUpdatedByServer = false;


    private Mesh[] meshesHead;

    private int indexHeadAnim = 0;
    private const float FPS_HEAD_ANIM = 2;
    private float timeHeadAnim = 0f;


    public void EnablePlayerInput()
    {
        isPlayerInputEnabled = true;
    }

    public void SetIsUpdatedByServer(bool isUpdatedByServer)
    {
        this.isUpdatedByServer = isUpdatedByServer;
    }

    //only to be called on one ball; the other balls will be destroyed automatically
    public void SafelyDestroyBall(Ball ball)
    {
        //Debug.Log("SafelyDestroyBall() on ball: " + ball.transform.name + " at frame " + Time.frameCount );
        if (HasBallConnectedToHead(ball))
        {
            Debug.LogWarning("This is not supposed to be called for the crane. Ball needs to be split off first using Snake.Split()");
        }
        else if (HasBall(ball))
        {
            bool found = false;
            int snakePartIndex = 0;
            while (!found && snakePartIndex < snakeParts.Count)
            {
                int snakeBallIndex = 0;
                while (!found && snakeBallIndex < snakeParts[snakePartIndex].snakeBalls.Count)
                {
                    if (snakeParts[snakePartIndex].snakeBalls[snakeBallIndex].ball == ball)
                    {
                        found = true;
                        snakeParts[snakePartIndex].Destroy();
                        snakeParts.RemoveAt(snakePartIndex);
                        //Debug.Log("Destroying SnakePart Ball: " + ball.transform.name + " at frame " + Time.frameCount );
                    }

                    ++snakeBallIndex;
                }
                ++snakePartIndex;
            }
        }
        else
        {
            GameObject.Destroy(ball.gameObject);
        }
    }

    private void Awake()
    {
        transform.localScale = new Vector3(Game.SNAKE_HEAD_DIAMETER, Game.SNAKE_HEAD_DIAMETER, Game.SNAKE_HEAD_DIAMETER);
        transform.name = "Snake";
        // GetComponent<Renderer>().material.color = Ball.GetRandomColor();
        // GetComponent<Renderer>().material.EnableKeyword("_EMISSION");
        // GetComponent<Renderer>().material.SetColor("_EmissionColor", Color.gray);

        GameObject head1 = Resources.Load<GameObject>("Prefabs/snake_head_1_mesh");
        meshesHead = new Mesh[]{
            head1.GetComponent<MeshFilter>().sharedMesh,
            Resources.Load<GameObject>("Prefabs/snake_head_2_mesh").GetComponent<MeshFilter>().sharedMesh,
            Resources.Load<GameObject>("Prefabs/snake_head_3_mesh").GetComponent<MeshFilter>().sharedMesh
        };
        GetComponent<Renderer>().material = head1.GetComponent<Renderer>().sharedMaterial;
    }


    void Start()
    {
        movementDirection = Quaternion.AngleAxis(Random.Range(0f, 360f), Vector3.up) * Vector3.forward;
        Snake.NumBallsDestroyedBySnake = 0;
    }

    void Update()
    {
        UpdateHeadAnimation();

        if (isUpdatedByServer)
        {
            return;
        }
        if (!isPlayerInputEnabled)
        {
            return;
        }


        FindCollisionsWithSelf();

        foreach (SnakeBall snakeBall in snakeBalls)
        {
            snakeBall.Update();
        }

        UpdateMovement();

        foreach (SnakePart snakePart in snakeParts)
        {
            snakePart.Update();
        }

    }

    private void UpdateHeadAnimation()
    {
        timeHeadAnim += Time.deltaTime;
        while (timeHeadAnim > (1f / FPS_HEAD_ANIM))
        {
            timeHeadAnim -= (1f / FPS_HEAD_ANIM);
            indexHeadAnim++;
            GetComponent<MeshFilter>().sharedMesh = GetHeadMeshByIndex(indexHeadAnim);
        }
    }

    private Mesh GetHeadMeshByIndex(int index)
    {
        return meshesHead[index% meshesHead.Length];
    }

    public void FindCollisionsWithSelf()
    {
        //verify whether any of the balls of the snake collided with the head
        bool collisionWithSelf = false;

        if (snakeBalls.Count > 3) //do not allow a snake to cut off its own body at the first three balls
        {
            int snakeBallIndex = 3;
            while (!collisionWithSelf && snakeBallIndex < snakeBalls.Count)
            {
                if (snakeBalls[snakeBallIndex].ball != null && snakeBalls[snakeBallIndex].IsGrown())
                {
                    float distance = (transform.position - snakeBalls[snakeBallIndex].ball.transform.position).magnitude;

                    if (distance < transform.localScale.x / 2f + snakeBalls[snakeBallIndex].ball.transform.localScale.x / 2f)
                    {
                        lastEdibleBallNearMouth = snakeBalls[snakeBallIndex].ball;
                        collisionWithSelf = true;
                    }
                }

                ++snakeBallIndex;
            }
        }
    }

    public void UpdateMovement()
    {
        transform.position += movementDirection * SPEED * Time.deltaTime;

        if (transform.position.x < -1f * Game.BALL_PIT_WIDTH / 2f)
        {
            transform.position = new Vector3(   -1f * Game.BALL_PIT_WIDTH / 2f + (-1f * Game.BALL_PIT_WIDTH / 2f - transform.position.x),
                                                transform.position.y,
                                                transform.position.z);
            movementDirection = new Vector3(-movementDirection.x, movementDirection.y, movementDirection.z);
        }
        else if (transform.position.x > Game.BALL_PIT_WIDTH / 2f)
        {
            transform.position = new Vector3(   Game.BALL_PIT_WIDTH / 2f - (transform.position.x - Game.BALL_PIT_WIDTH / 2f),
                                                transform.position.y,
                                                transform.position.z);
            movementDirection = new Vector3(-movementDirection.x, movementDirection.y, movementDirection.z);
        }

        if (transform.position.z < -1f * Game.BALL_PIT_DEPTH / 2f)
        {
            transform.position = new Vector3( transform.position.x,
                                                transform.position.y,
                                                -1f * Game.BALL_PIT_DEPTH / 2f + (-1f * Game.BALL_PIT_DEPTH / 2f - transform.position.z));
            movementDirection = new Vector3(movementDirection.x, movementDirection.y, -movementDirection.z);
        }
        else if (transform.position.z > Game.BALL_PIT_DEPTH / 2f)
        {
            transform.position = new Vector3( transform.position.x,
                                                transform.position.y,
                                                Game.BALL_PIT_DEPTH / 2f - (transform.position.z - Game.BALL_PIT_DEPTH / 2f));
            movementDirection = new Vector3(movementDirection.x, movementDirection.y, -movementDirection.z);
        }
        SetNewMovementDirectionIfClicked();

        transform.forward = movementDirection;
    }

    //returns true if the ball should be counted towards the score
    public bool EatBall(Ball ball)
    {
        bool ballToBeCountedTowardsScore = true;
        bool isFirstBall = (snakeBalls.Count == 0);
        bool ballConnectedToHead = false;

        ball.EnableSnakeColor(true);

        Transform ballParent;
        if (isFirstBall)
        {
            ballParent = GameObject.FindObjectOfType<Snake>().transform;
        }
        else
        {
            ballParent = snakeBalls[snakeBalls.Count - 1].ball.transform;

            if (HasBallConnectedToHead(ball))
            {
                ballConnectedToHead = true;

                //NOTE this also works when it's the last ball (+ stuff gets destroyed anyway)
                Split(ball, null);
                snakeParts[snakeParts.Count - 1].Destroy();
                snakeParts.RemoveAt(snakeParts.Count - 1);
                ballToBeCountedTowardsScore = false;
            }
        }

        if (!ballConnectedToHead)
        {
            SnakeBall snakeBall = new SnakeBall(ball, ballParent);
            snakeBalls.Add(snakeBall);
        }

        if (ball == lastEdibleBallNearMouth)
        {
            lastEdibleBallNearMouth = null;
        }

        return ballToBeCountedTowardsScore;
    }

    public Ball GetLastEdibleBallNearMouth()
    {
        return lastEdibleBallNearMouth;
    }

    public bool HasBall(Ball ball)
    {
        bool found = HasBallConnectedToHead(ball);

        int snakePartsIndex = 0;
        while(!found && snakePartsIndex < snakeParts.Count)
        {
            int snakeBallIndex = 0;
            while (!found && snakeBallIndex < snakeParts[snakePartsIndex].snakeBalls.Count)
            {
                if (snakeParts[snakePartsIndex].snakeBalls[snakeBallIndex].ball == ball)
                {
                    found = true;
                }

                ++snakeBallIndex;
            }
            ++snakePartsIndex;
        }

        return found;
    }

    public bool HasBallConnectedToHead(Ball ball)
    {
        bool found = false;
        int snakeBallIndex = 0;
        while (!found && snakeBallIndex < snakeBalls.Count)
        {
            if (snakeBalls[snakeBallIndex].ball == ball)
            {
                found = true;
            }

            ++snakeBallIndex;
        }

        return found;
    }

    //NOTE this method has not been tested yet, though relies largely on another method. @Jerry: let me know if you run into any issue
    public void Split(Ball ball, Transform newParent)
    {
        if (snakeBalls.Count > 0)
        {
            bool found = false;
            int snakeBallIndex = 0;
            int foundSnakeBallIndex = -1;
            while (!found && snakeBallIndex < snakeBalls.Count)
            {
                if (snakeBalls[snakeBallIndex].ball == ball)
                {
                    found = true;
                    foundSnakeBallIndex = snakeBallIndex;
                }

                ++snakeBallIndex;
            }

            if (found)
            {
                Split(foundSnakeBallIndex, newParent);
            }
        }
        else
        {
            Debug.LogWarning("Snake.Split called but there are no balls in snakeBalls.");
        }
    }

    public void Split(int snakeBallIndex, Transform newParent)
    {
        //TODO remove after the other Split method has been fully tested
        //Debug.Log("Split called with snakeBallIndex " + snakeBallIndex + " and " + snakeBalls.Count + " elements in snakeBalls");

        if (snakeBallIndex < snakeBalls.Count)
        {
            List<SnakeBall> tailEndSnakeBalls = new List<SnakeBall>();

            for (int i = snakeBallIndex; i < snakeBalls.Count; ++i)
            {
                tailEndSnakeBalls.Add(snakeBalls[i]);
            }

            for (int i = 0; i < tailEndSnakeBalls.Count; ++i)
            {
                snakeBalls.RemoveAt(snakeBalls.Count - 1);
            }

            snakeParts.Add(new SnakePart(tailEndSnakeBalls, newParent));

            //TODO remove after the other Split method has been fully tested
            //Debug.Log("After the split: " + snakeBalls.Count + " elements in snakeBalls and " + tailEndSnakeBalls.Count + " elements in tailEndSnakeBalls.");
        }
        else
        {
            Debug.LogWarning("Invalid index to split the snake on");
        }
    }

    //NOTE: this is not called when running into yourself because the balls that are part of the Snake are set to kinematic
    private void OnCollisionEnter(Collision other)
    {
        Ball ball = other.gameObject.GetComponent<Ball>();
        if (ball)
        {
            if (ball.IsEdible())
            {
                lastEdibleBallNearMouth = ball;
            }
        }
    }

    // this is called when balls drop below the ground and are destroyed
    public void RemoveBallIfPresent(Ball ball)
    {
        if (snakeBalls.Count > 0)
        {
            bool found = false;
            int snakeBallIndex = snakeBalls.Count - 1;

            if (snakeBalls[snakeBallIndex].ball == ball)
            {
                found = true;
                snakeBalls.RemoveAt(snakeBallIndex);
            }
            --snakeBallIndex;

            while (!found && snakeBallIndex >= 0)
            {
                if (snakeBalls[snakeBallIndex].ball == ball)
                {
                    found = true;

                    if (snakeBallIndex > 0)
                    {
                        snakeBalls[snakeBallIndex + 1].parent = snakeBalls[snakeBallIndex - 1].ball.transform;
                    }
                    else
                    {
                        snakeBalls[snakeBallIndex + 1].parent = transform;
                    }

                    snakeBalls.RemoveAt(snakeBallIndex);
                }

                --snakeBallIndex;
            }
        }       
    }

    public void SetNewMovementDirectionIfClicked()
    {
        if (Input.GetMouseButton(0))
        {
            Plane XZ = new Plane(Vector3.up, Vector3.zero);

            Ray ray = Camera.main.ScreenPointToRay(Input.mousePosition);
            float enter = 0.0f;
            if (XZ.Raycast(ray, out enter))
            {
                Vector3 hitPoint = ray.GetPoint(enter);

                Vector3 newDirection = hitPoint - new Vector3(transform.position.x, 0f, transform.position.z);
                if (newDirection.sqrMagnitude > 0.05f * 0.05f)
                {
                    movementDirection = newDirection.normalized;
                }
            }
        }

    }
    

}
